import pandas as pd
import numpy as np
import pytest
import unittest
import sys
import context

from fastbt.utils import *

def equation(a,b,c,x,y):
	return a*x**2 + b*y + c

def test_multiargs_simple():
	seq = pd.Series([equation(1,2,3,4,y) for y in range(20, 30)]).sort_index()
	seq.index = range(20,30)
	constants = {'a':1, 'b':2, 'c':3, 'x':4}
	variables = {'y': range(20, 30)}
	par = multi_args(equation, constants=constants, variables=variables).sort_index()
	# Check both values and indexes
	for x,y in zip(seq, par):
		assert x == y
	for x,y in zip (seq.index, par.index):
		assert (x,) == y

def test_multiargs_product():
	seq = []
	for x in range(0,10):
		for y in range(10,15):
			seq.append(equation(1,2,3,x,y))
	index = pd.MultiIndex.from_product([range(0, 10), range(10, 15)])
	seq = pd.Series(seq)
	seq.index = index
	seq = seq.sort_index()
	constants =  {'a':1, 'b':2, 'c':3}
	variables = {'x': range(0, 10), 'y': range(10,15)}
	par = multi_args(equation, constants=constants, 
		variables=variables, isProduct=True).sort_index()
	# Check both values and indexes
	for x,y in zip(seq, par):
		assert x == y
	for x,y in zip (seq.index, par.index):
		assert x == y	

def test_multiargs_max_limit():
	seq = []
	for x in range(0,100):
		for y in range(100, 150):
			seq.append(equation(1,2,3,x,y))
	index = pd.MultiIndex.from_product([range(0, 100), range(100, 150)])
	seq = pd.Series(seq)
	seq.index = index
	seq = seq.sort_index()
	constants =  {'a':1, 'b':2, 'c':3}
	variables = {'x': range(0, 100), 'y': range(100,150)}
	par = multi_args(equation, constants=constants, 
		variables=variables, isProduct=True).sort_index()
	assert len(par) == 1000
	assert len(seq) == 5000
	# Check both values and indexes
	for x,y in zip(seq, par):
		assert x == y
	for x,y in zip (seq.index, par.index):
		assert x == y

@pytest.mark.parametrize("maxLimit", [2000, 3000, 5000, 10000])
def test_multiargs_max_limit_adjust(maxLimit):
	seq = []
	for x in range(0,100):
		for y in range(100, 150):
			seq.append(equation(1,2,3,x,y))
	index = pd.MultiIndex.from_product([range(0, 100), range(100, 150)])
	seq = pd.Series(seq)
	seq.index = index
	seq = seq.sort_index()
	constants =  {'a':1, 'b':2, 'c':3}
	variables = {'x': range(0, 100), 'y': range(100,150)}
	par = multi_args(equation, constants=constants, 
		variables=variables, isProduct=True, maxLimit=maxLimit).sort_index()
	assert len(par) == min(maxLimit, 5000)
	assert len(seq) == 5000
	# Check both values and indexes
	for x,y in zip(seq, par):
		assert x == y
	for x,y in zip (seq.index, par.index):
		assert x == y

def test_tick():
	assert tick(112.71) == 112.7
	assert tick(112.73) == 112.75
	assert tick(1054.85, tick_size=0.1) == 1054.8
	assert tick(1054.851, tick_size=0.1) == 1054.9
	assert tick(104.73, 1) == 105
	assert tick(103.2856, 0.01) == 103.29
	assert tick(0.007814, 0.001) == 0.008
	assert tick(0.00003562, 0.000001) == 0.000036
	assert tick(0.000035617, 0.00000002) == 0.00003562

def test_tick_series():
	s = pd.Series([100.43, 200.32, 300.32])
	result = [100.45, 200.3, 300.3]
	for x,y in zip(tick(s), result):
		assert x==y

def test_stop_loss():
	assert stop_loss(100, 3) == 97
	assert stop_loss(100, 3, order='S') == 103
	assert stop_loss(1013, 2.5, order='B', tick_size=0.1) == 987.7
	assert stop_loss(100, -3) == 103 # This should be depreceated
	assert stop_loss(100, -3, order='S') == 97

def test_stop_loss_error():
	with pytest.raises(ValueError):
		assert stop_loss(100, 3, 'BS')

def test_stop_loss_series():
	p = pd.Series([100.75, 150.63, 180.32])
	result = [95.71, 143.1, 171.3]
	for x,y in zip(stop_loss(p, 5, tick_size=0.01), result):
		assert pytest.approx(x, rel=0.001, abs=0.001) == y

	# Test for sell
	result = [105.79, 158.16, 189.34]
	for x,y in zip(stop_loss(p, 5, order='S', tick_size=0.01), result):
		assert pytest.approx(x, rel=0.001, abs=0.001) == y

def test_create_orders_simple():
	df = pd.DataFrame(np.arange(20).reshape(5,4), columns=list('ABCD'))
	orders = create_orders(df, {'A': 'one', 'B': 'two', 'C': 'three', 'D': 'four'},
		exchange='NSE', num=range(5))
	df['exchange'] = 'NSE'
	df['num'] = [0,1,2,3,4]
	assert list(orders.columns) == ['one', 'two', 'three', 'four', 'exchange', 'num']
	assert list(df.exchange) == ['NSE'] * 5

class TestRecursiveMerge(unittest.TestCase):

	def setUp(self):
		df1 = pd.DataFrame(np.random.randn(6,3), columns=list('ABC'))
		df2 = pd.DataFrame(np.random.randn(10,3), columns=list('DEF'))
		df3 = pd.DataFrame(np.random.randn(7,4), columns=list('GHIJ'))
		df4 = pd.DataFrame(np.random.randn(10,7), columns=list('AMNDXYZ'))
		df1['idx'] = range(100,106)
		df2['idx'] = range(100, 110)
		df3['idx'] = range(100, 107)
		df4['idx'] = range(100, 110)
		self.dfs = [df1, df2, df3, df4]

	def test_recursive_merge_simple(self):
		df = recursive_merge(self.dfs)
		assert len(df) == 6
		assert df.shape == (6, 21)
		assert df.loc[3, 'X'] == self.dfs[3].loc[3, 'X']
		assert df.iloc[2, 11] == self.dfs[2].iloc[2, 3]

	def test_recursive_on(self):
		df = recursive_merge(self.dfs, on=['idx'])
		assert df.shape == (6, 18)
		assert df.loc[3, 'X'] == self.dfs[3].loc[3, 'X']
		assert df.iloc[2, 11] == self.dfs[3].iloc[2, 0]

	def test_recursive_on(self):
		dct = {'1': 'D', '2': 'G', '3': 'X'}
		df = recursive_merge(self.dfs, on=['idx'], columns=dct)
		assert df.shape == (6, 7)
		assert list(sorted(df.columns)) == ['A', 'B', 'C', 'D', 'G', 'X', 'idx']
		assert df.loc[3, 'X'] == self.dfs[3].loc[3, 'X']


import unittest
import pandas as pd
import sys
from sqlalchemy import create_engine
import pytest
from functools import partial
from random import randint
import yaml
import os
import context

from fastbt.rapid import *
R = lambda x: round(x, 2)

def compare(frame1, frame2):
    """
    Compare a random value from 2 dataframes
    return
        True if values are equal else False
    """
    r1 = randint(0, len(frame1)-1)
    r2 = randint(0, len(frame1.columns) - 1)
    return frame1.iloc[r1, r2] == frame2.iloc[r1, r2]


class TestRapidFetchData(unittest.TestCase):

    def setUp(self):
        self.con = create_engine('sqlite:///tests/data/data.sqlite3')
        self.tbl = 'eod'

    def test_fetch_data(self):
        """
        Additional date to be included for correct results due to
        """
        universe = ['one', 'two']
        start, end = '2018-01-01 00:00:00.000000', '2018-01-02 00:00:00.000000'
        data = fetch_data(universe, start, end, self.con, self.tbl)
        self.assertEqual(data.shape[0], 4)

        universe = ['one', 'two', 'six']
        data = fetch_data(universe, start, end, self.con, self.tbl)
        self.assertEqual(data.shape[0], 6)

        end = '2018-01-06 00:00:00.000000'
        universe.append('five')
        data = fetch_data(universe, start, end, self.con, self.tbl)
        self.assertEqual(data.shape[0], 24)

    def test_fetch_data_condition(self):
        universe = ['one', 'two', 'three', 'four', 'five', 'six']
        start, end = '2018-01-01 00:00:00.000000', '2018-01-06 00:00:00.000000'
        condition = ['open > 100']
        data = fetch_data(universe, start, end, self.con, self.tbl,
                where_clause = condition)       
        self.assertEqual(data.shape[0], 17)
        condition.append('volume > 200000')
        data = fetch_data(universe, start, end, self.con, self.tbl,
                where_clause = condition)       
        self.assertEqual(data.shape[0], 12)

class TestRapidPrepareData(unittest.TestCase):

    #TO DO: return in case of Empty dataframe
    def setUp(self):
        from sqlalchemy import create_engine
        con = create_engine('sqlite:///tests/data/data.sqlite3')
        tbl = 'eod'
        universe = ['one', 'two', 'three', 'four', 'five', 'six']
        start, end = '2018-01-01 00:00:00.000000', '2018-01-06 00:00:00.000000'
        self.data = fetch_data(universe, start, end, con, tbl)

    def test_prepare_data(self):
        self.assertEqual(self.data.shape[0], 36)
        self.assertEqual(self.data.shape[1], 8)

        columns = [
            {'F': {'formula': '(open+close)/2', 'col_name': 'avgprice'}},
            {'I': {'indicator': 'SMA', 'period': 3, 'lag': 1, 'col_name': 'SMA3'}}
        ]
        conditions = [
            {'F': {'formula': 'open > prevclose', 'col_name': 'sig1'}},
            {'F': {'formula': 'open < sma3', 'col_name': 'sig2'}}
        ]
        data = prepare_data(self.data, columns)
        self.assertEqual(data.shape[1], 10)
        data = prepare_data(data, conditions)
        self.assertEqual(data.shape[1], 12)
        self.assertEqual(data.query('sig1==1').shape[0], 20)
        self.assertEqual(data.query('sig2==1').shape[0], 5) 

class TestRapidApplyPrices(unittest.TestCase):

    def setUp(self):
        from sqlalchemy import create_engine
        con = create_engine('sqlite:///tests/data/data.sqlite3')
        tbl = 'eod'
        universe = ['one', 'two', 'three', 'four', 'five', 'six']
        start, end = '2018-01-01 00:00:00.000000', '2018-01-06 00:00:00.000000'
        self.data = fetch_data(universe, start, end, con, tbl)

    def test_apply_price_buy(self):
        # Simple default parameters
        idx = pd.IndexSlice
        R = lambda x: round(x, 2)
        conditions = ['open > prevclose']
        df = apply_prices(self.data, conditions, 'open', 3, 'B')
        self.assertEqual(df.shape[0], 20)
        df.set_index(['timestamp', 'symbol'], inplace=True)
        self.assertEqual(df.at[idx['2018-01-06', 'one'], 'price'], 10.65)
        self.assertEqual(df.query('low <= stop_loss').shape[0], 7)
        self.assertEqual(df.query('low <= stop_loss').price.sum(), 562.45)
        self.assertEqual(df.at[idx['2018-01-05', 'four'], 'sell'], 169.25)
        self.assertEqual(df.at[idx['2018-01-05', 'six'], 'sell'], 63.45)
        self.assertEqual(R(df.at[idx['2018-01-05', 'five'], 'stop_loss']), 25.70)
        self.assertEqual(df.at[idx['2018-01-05', 'five'], 'sell'], 27.4)

    def test_apply_price_sell(self):
        idx = pd.IndexSlice
        R = lambda x: round(x, 2)
        conditions = ['open > prevclose']
        df = apply_prices(self.data, conditions, 'open', 3, 'S')
        df.set_index(['timestamp', 'symbol'], inplace=True)
        self.assertEqual(df.query('stop_loss >= high').shape[0], 13)
        self.assertEqual(R(df.at[idx['2018-01-02', 'three'], 'stop_loss']), 105.05)
        self.assertEqual(df.at[idx['2018-01-05', 'four'], 'price'], 174.5)
        self.assertEqual(df.at[idx['2018-01-03', 'three'], 'buy'], 110) 

    def test_order_raise_error(self):
        pass

class TestRapidRunStrategy(unittest.TestCase):

    def setUp(self):
        from sqlalchemy import create_engine
        con = create_engine('sqlite:///tests/data/data.sqlite3')
        tbl = 'eod'
        universe = ['one', 'two', 'three', 'four', 'five', 'six']
        start, end = '2018-01-01 00:00:00.000000', '2018-01-06 00:00:00.000000'
        data = fetch_data(universe, start, end, con, tbl)
        conditions = ['open > prevclose']
        self.data = apply_prices(data, conditions, 'open', 3, 'B')

    def test_run_strategy_default(self):
        idx = pd.IndexSlice
        df = run_strategy(self.data, 'price', True, 5)
        df = df.set_index(['timestamp', 'symbol']).sort_index()
        print('IDX', df.loc[idx['2018-01-03', 'four'], 'stop_loss'])
        self.assertEqual(R(df.at[idx['2018-01-03', 'four'], 'stop_loss']), 162.95)
        self.assertEqual(R(df.at[idx['2018-01-05', 'five'], 'price']), 26.5)


    def test_run_strategy_custom(self):
        def func(x):
            return x.iloc[[0]]
        df = run_strategy(self.data, 'price', True, 5, func)
        self.assertEqual(len(df), 6)

class TestRapidGetOutput(unittest.TestCase):

    def setUp(self):
        from sqlalchemy import create_engine
        con = create_engine('sqlite:///tests/data/data.sqlite3')
        tbl = 'eod'
        universe = ['one', 'two', 'three', 'four', 'five', 'six']
        start, end = '2018-01-01 00:00:00.000000', '2018-01-06 00:00:00.000000'
        data = fetch_data(universe, start, end, con, tbl)
        conditions = ['open > prevclose']
        data = apply_prices(data, conditions, 'open', 3, 'B')
        self.data = run_strategy(data, 'price', True, 5)

    def test_get_output_default(self):
        result = get_output(self.data, 100000, 1, 0,0)
        self.assertEqual(R(result.profit.sum()), -715.09)
        self.assertEqual(result.qty.sum(), 26506)
        by_day = result.groupby('timestamp').profit.sum()
        self.assertEqual(R(by_day.loc['2018-01-04']), -3153.15)

def test_backtest_data():
    import yaml
    with open('tests/data/backtest.yaml') as f:
        kwargs = yaml.load(f)
    kwargs['connection'] = create_engine('sqlite:///tests/data/data.sqlite3')
    kwargs['tablename'] = 'eod'
    result_one = backtest(**kwargs)

    data = pd.read_csv('tests/data/sample.csv', parse_dates=['timestamp'])
    del kwargs['connection']
    del kwargs['tablename']
    kwargs['data'] = data
    result_two = backtest(**kwargs)
    assert metrics(result_one, 100000) == metrics(result_two, 100000)
    assert result_one.shape == result_two.shape
    from random import randint
    for i in range(10):
        assert compare(result_one, result_two)

def test_stop_loss_zero():
    pass

def _build_input_output():
    """
    This function builds the input and output
    from results for passing to pytest
    """
    data = pd.read_csv('tests/data/results.csv').to_dict(orient='records')
    with open('tests/data/BT.yaml') as f:
        params = yaml.load(f)

    input_map = {
    'start': 'start',
    'end': 'end',
    'capital': 'capital',
    'lev': 'leverage',
    'comm': 'commission',
    'slip': 'slippage',
    'sl': 'stop_loss',
    'order': 'order',
    'limit': 'limit',
    'universe': 'universe',
    'sort_by': 'sort_by',
    'sort_mode': 'sort_mode'
    }
    output_cols = ['profit', 'commission', 'slippage', 'net_profit',
                    'high', 'low', 'drawdown', 'returns', 'sharpe']
    final = []
    for d in data:
        kwargs = {}
        for k,v in input_map.items():
            kwargs[v] = d[k]
        if kwargs.get('universe') != 'all':
            kwargs['universe'] = kwargs['universe'].split(',')
        kwargs['start'] = pd.to_datetime(kwargs['start'])
        kwargs['end'] = pd.to_datetime(kwargs['end'])
        kwargs.update(params.get(d['strategy']))
        result = {}
        for col in output_cols:
            result[col] = d[col]
        final.append((kwargs, result))
    return final

con = create_engine('sqlite:///tests/data/data.sqlite3')
tbl = 'eod'
bt = partial(backtest, connection=con, tablename=tbl)
@pytest.mark.parametrize("kwargs, expected", _build_input_output())
def test_backtest_results(kwargs, expected):    
    result = metrics(bt(**kwargs), kwargs['capital'])
    for k,v in result.items():
        assert pytest.approx(v, rel=0.001, abs=0.001) == expected[k]

def test_empty_dataframe_result():
    params = {
        'start': '2020-01-01',
        'end': '2020-01-05',
        'sort_by': 'open'
    }
    with pytest.raises(ValueError):
        bt(**params)

    params.update({'start': '2017-01-01', 'conditions':['open > 20000']})
    with pytest.raises(ValueError):
        bt(**params)

def test_no_datasource():
    params = {
        'start': '2018-01-01',
        'end': '2020-01-05'
    }
    with pytest.raises(ValueError):
        backtest(**params)
  
    params.update({'tablename': tbl})
    with pytest.raises(ValueError):
        backtest(**params)   
    del params['tablename']

    params.update({'connection': con})
    with pytest.raises(ValueError):
        backtest(**params)

    params.update({'tablename': tbl})
    backtest(**params)

def test_no_columns():
    params = {
        'start': '2018-01-01',
        'end': '2018-01-07',
        'sort_by': 'open',
        'conditions': ['open > 0'],
        'limit': 10
    }
    assert len(bt(**params)) == 36

def test_no_conditions():
    params = {
        'start': '2018-01-01',
        'end': '2018-01-07',
        'sort_by': 'open',
        'columns': [{'F': {'formula': '(open+close)/2', 'col_name': 'avgprice'}}],
        'limit': 10
    }
    assert len(bt(**params)) == 36

def test_no_columns_no_conditions():
    params = {
        'start': '2018-01-01',
        'end': '2018-01-07',
        'sort_by': 'open',
        'limit': 10
        }
    df1 = pd.read_csv('tests/data/sample.csv', parse_dates=['timestamp'])
    df1 = df1.sort_values(by=['timestamp', 'symbol'])
    df2 = bt(**params).sort_values(by=['timestamp', 'symbol'])
    for i in range(20):
        assert compare(df1, df2)

def test_strategy_output():
    params = {
        'start': '2018-01-01',
        'end': '2018-01-07',
        'sort_by': 'price',
        'limit': 10
    }
    df1 = pd.read_csv('tests/data/sample.csv', parse_dates=['timestamp'])
    df1 = df1.sort_values(by=['timestamp', 'symbol'])
    strategy = lambda x: x
    output = lambda x:x
    df2 = bt(**params, strategy=strategy, output=output).sort_values(by=['timestamp', 'symbol'])
    for i in range(20):
        assert compare(df1, df2)

@pytest.mark.parametrize("stop_loss", [1,2,3])
@pytest.mark.parametrize("order", ['B', 'S'])
@pytest.mark.parametrize("limit", [3,5])
def test_same_results(stop_loss, order, limit):
    # test whether results are same for both backtest
    # function and when functions are run individually
    params = {
    'start': '2018-01-01',
    'end': '2018-01-10',
    'price': 'open * 0.999',
    'columns':[{'F': {'formula': '(open+close)/2', 'col_name': 'avgprice'}}],
    'conditions': ['open > 50'],
    'sort_by': 'price'
    }
    params.update({'stop_loss': stop_loss, 'order': order, 'limit': limit})
    result_one = bt(**params)
    df = fetch_data('all', '2018-01-01','2018-01-10', con, tbl) 
    df = prepare_data(df, params['columns'])
    df = apply_prices(df, params['conditions'], params['price'],
        stop_loss=stop_loss, order=order)
    df = run_strategy(df, params['sort_by'], True, limit=limit)
    result_two = get_output(df, 100000, 1)
    for i in range(50):
        assert compare(result_one, result_two)

def test_backtest_from_excel():
    my_result = {
        'profit': -715.09,
        'commission': 0.0,
        'slippage': 0.0,
        'net_profit': -715.09,
        'high': 1200.76,
        'low': -2599.34,
        'drawdown': -0.026,
        'returns': -0.007,
        'sharpe': -0.065
    }
    con = create_engine('sqlite:///tests/data/data.sqlite3')
    tbl = 'eod'
    result = backtest_from_excel('tests/data/backtest.xls', connection=con, tablename=tbl)
    result = metrics(result, 100000)
    for k,v in my_result.items():
        assert round(result[k], 3) == v

def test_backtest_from_json():
    my_result = {
        'profit': -715.09,
        'commission': 0.0,
        'slippage': 0.0,
        'net_profit': -715.09,
        'high': 1200.76,
        'low': -2599.34,
        'drawdown': -0.026,
        'returns': -0.007,
        'sharpe': -0.065
    }
    con = create_engine('sqlite:///tests/data/data.sqlite3')
    tbl = 'eod'
    result = backtest_from_json('tests/data/backtest.json', connection=con, tablename=tbl)
    result = metrics(result, 100000)
    for k,v in my_result.items():
        assert round(result[k], 3) == v

def test_backtest_from_yaml():
    my_result = {
        'profit': -715.09,
        'commission': 0.0,
        'slippage': 0.0,
        'net_profit': -715.09,
        'high': 1200.76,
        'low': -2599.34,
        'drawdown': -0.026,
        'returns': -0.007,
        'sharpe': -0.065
    }
    con = create_engine('sqlite:///tests/data/data.sqlite3')
    tbl = 'eod'
    result = backtest_from_yaml('tests/data/backtest.yaml', connection=con, tablename=tbl)
    result = metrics(result, 100000)
    for k,v in my_result.items():
        assert round(result[k], 3) == v



file_patterns = {
    'bhav': (
        'https://www.nseindia.com/content/historical/EQUITIES/{year}/{month}/cm{date}bhav.csv.zip',
        lambda x: {
            'year': x.year,
            'month': x.strftime('%b').upper(),
            'date': x.strftime('%d%b%Y').upper()            
        }
    ),
    
    'sec_del': (
        'https://www.nseindia.com/archives/equities/mto/MTO_{date}.DAT',
        lambda x: {
            'date': x.strftime('%d%m%Y')
        }
    ),
    
    'bhav_pr': (
        'https://www.nseindia.com/archives/equities/bhavcopy/pr/PR{date}.zip',
        lambda x: {
            'date': x.strftime('%d%m%Y')
        }
    ),
    
    'derivatives': (
        'https://www.nseindia.com/content/historical/DERIVATIVES/{year}/{month}/fo{date}bhav.csv.zip',
        lambda x: {
            'year': x.year,
            'month': x.strftime('%b').upper(),
            'date': x.strftime('%d%b%Y').upper()           
        }
    )
}




"""
Load data to database
"""

import os
import pandas as pd

def apply_adjustment(df, adj_date, adj_value,
                    adj_type='mul',date_col='date',
                    cols=['open','high', 'low', 'close']):
    """
    Apply adjustment to a given stock
    df
        dataframe of the given stock
    adj_date
        date from which the adjustment is
        to be made
    adj_value
        value to be adjusted
    adj_type
        method of adjustment **mul/sub**
        mul means multiply all the values
            such as splits and bonuses
        sub means subtract the values
            such as dividends
 
    date_col
        date column  on which the adjustment
        is to be applied
    cols
        columns to which the adjustment is to
        be made

    Notes
    -----
    1) You can use negative values to add to the
    stock value by using **adj_type=sub**
    2) Adjustment is applied prior to all dates
    in the dataframe
    3) In case your dataframe has date or
    symbol as indexes, reset them
    """
    df = df.set_index(date_col).sort_index()
    values_on_adj_date = df.loc[adj_date, cols].copy()
    if adj_type == "mul":
        adjusted_values = (df.loc[:adj_date, cols] * adj_value).round(2)
    elif adj_type == "sub":
        adjusted_values = (df.loc[:adj_date, cols] - adj_value).round(2)
    else:
        raise ValueError('adj_type should be either mul or sub')
    df.loc[:adj_date, cols] = adjusted_values
    df.loc[adj_date, cols] = values_on_adj_date
    return df.reset_index()    

class DataLoader(object):
    """
    Data Loader class
    """
    def __init__(self, directory, mode='HDF', engine=None, 
                 tablename=None):
        """
        Initialize parameters
        directory
            directory to search files
        mode
            HDF/SQL - should be explicitly specified
        engine
            filename in case of HDF
            SQL Alchemy connection string in case of engine
        tablename
            table where data is to be written
        parse dates
            list of columns to be parsed as date
        """
        if mode not in ['SQL', 'HDF']:
            raise TypeError('No mode specified; should be HDF or SQL')
        self.directory = directory
        self.mode = mode
        self.engine = engine
        self.tablename = tablename
        
    def _initialize_HDF_file(self):
        import hashlib
        hash = hashlib.sha1().hexdigest()
        with pd.HDFStore(self.engine) as store:
            s = pd.Series(['hash'*2])            
            if len(store.keys()) == 0:
                store.append('updated/'+self.tablename, s)

    def _write_to_HDF(self, **kwargs):
        """
        Write data to HDF file
        """
        update_table = '/updated/' + self.tablename
        data_table = '/data/' + self.tablename
        updated_list = []
        with pd.HDFStore(self.engine) as store:
            if update_table in store.keys():
                updated_list = store.get(update_table).values

        if kwargs.get('columns'):
            columns = kwargs.pop('columns')
        else:
            columns = None

        if kwargs.get('parse_dates'):
            parse_dates = kwargs.get('parse_dates')
        else:
            parse_dates = None

        if kwargs.get('postfunc'):
            postfunc = kwargs.pop('postfunc')
        else:
            postfunc = None


        # Iterating over the files
        for root, direc, files in os.walk(self.directory):
            for file in files:
                if file not in updated_list:
                    filename = os.path.join(root, file)
                    df = pd.read_csv(filename, **kwargs)       
                    df = df.rename(str.lower, axis='columns')
                    if columns:
                        df = df.rename(columns, axis='columns')
                    if not(parse_dates):
                        date_cols = ['date', 'time', 'datetime', 'timestamp']
                        for c in df.columns:
                            if c in date_cols:
                                df[c] = pd.to_datetime(df[c])
                    if postfunc:
                        df = postfunc(df, file, root)
                    df.to_hdf(self.engine, key=data_table, format='table',
                        append=True, data_columns=True)
                    # Updating the file data
                    pd.Series([file]).to_hdf(self.engine, key=update_table,
                        format='table', append=True)

    def _write_to_SQL(self, **kwargs):
        """
        Write data to SQL database     
        """
        update_table = 'updated_' + self.tablename
        data_table = self.tablename
        updated_list = []
        if self.engine.has_table(update_table):
            updated_list = pd.read_sql_table(update_table, self.engine).values

        if kwargs.get('columns'):
            columns = kwargs.pop('columns')
        else:
            columns = None

        if kwargs.get('parse_dates'):
            parse_dates = kwargs.get('parse_dates')
        else:
            parse_dates = None

        if kwargs.get('postfunc'):
            postfunc = kwargs.pop('postfunc')
        else:
            postfunc = None

        # Iterating over the files
        for root, direc, files in os.walk(self.directory):
            for file in files:
                if file not in updated_list:
                    filename = os.path.join(root, file)
                    df = pd.read_csv(filename, **kwargs)
                    df = df.rename(str.lower, axis='columns')
                    if columns:
                        df = df.rename(columns, axis='columns')
                    if not(parse_dates):
                        date_cols = ['date', 'time', 'datetime', 'timestamp']
                        for c in df.columns:
                            if c in date_cols:
                                df[c] = pd.to_datetime(df[c])
                    if postfunc:
                        df = postfunc(df, file, root)
                    s = pd.Series([file])
                    df.to_sql(data_table, con=self.engine, if_exists='append',
                    index=False, chunksize=1500)
                    # Updating the file data
                    s.to_sql(update_table, con=self.engine, if_exists='append', 
                        index=False, chunksize=1500)

    def load_data(self, **kwargs):
        """
        Load data into database
        kwargs
        columns
            column names as dictionary
            with key being column name from file
            and value being the column to be renamed
            ```
            {'OPENING': 'open', 'CLOSING': 'close'}
            ```
        parse_dates
            columns to be parsed as list
            If not given, any column with name
            date, datetime, time, timestamp is 
            automatically parse
        postfunc
            function to be run after reading the csv file  
        kwargs
            Any other arguments to the pandas read_csv function          
        """
        if self.mode == 'HDF':
            self._write_to_HDF(**kwargs)
        else:
            self._write_to_SQL(**kwargs)

    def apply_splits(self, directory='adjustments', 
        filename='splits.csv', symbol='symbol', timestamp='date'):
        """
        Apply splits recursively
        By default, only open, high, low, close and volume columns
        are modified
        """

        filename = os.path.join(directory, filename)
        try:
            splits = pd.read_csv(filename, parse_dates=[timestamp])
        except Exception as e:
            print(e)

        if self.mode == 'SQL':
            df = pd.read_sql_table(self.tablename, self.engine)
            for i, row in splits.iterrows():
                q = 'symbol == "{sym}"'
                temp = df.query(q.format(sym=row.at[symbol]))
                params = {
                    'adj_date': row.at[timestamp],
                    'adj_value': row.at['from']/row.at['to'],
                    'adj_type': 'mul',
                    'date_col': timestamp,
                    'cols': ['open', 'high', 'low', 'close']
                }
                temp = apply_adjustment(temp, **params)
                params.update({
                    'adj_value': row.at['to'] / row.at['from'],
                    'cols': ['volume']
                    })
                temp = apply_adjustment(temp, **params)
                cols = ['open', 'high', 'low', 'close', 'volume']
                temp.index = df.loc[df[symbol] == row.at[symbol]].index
                df.loc[temp.index] = temp
            df.to_sql(self.tablename, self.engine, if_exists='replace', index=False)
        elif self.mode == 'HDF':          
            df = pd.read_hdf(self.engine, '/data/'+ self.tablename)
            df.index = range(len(df))
            for i, row in splits.iterrows():
                q = 'symbol == "{sym}"'
                temp = df.query(q.format(sym=row.at[symbol]))
                params = {
                    'adj_date': row.at[timestamp],
                    'adj_value': row.at['from']/row.at['to'],
                    'adj_type': 'mul',
                    'date_col': timestamp,
                    'cols': ['open', 'high', 'low', 'close']
                }
                temp = apply_adjustment(temp, **params)
                params.update({
                    'adj_value': row.at['to'] / row.at['from'],
                    'cols': ['volume']
                    })
                temp = apply_adjustment(temp, **params)
                cols = ['open', 'high', 'low', 'close', 'volume']
                temp.index = df.loc[df[symbol] == row.at[symbol]].index  
                df.loc[temp.index] = temp
            df.to_hdf(self.engine, key='/data/'+self.tablename, format='table',
                        data_columns=True)


def collate_data(directory, function=None, **kwargs):
    """
    Given a directory of csv files with similar structure,
    create a dataframe by concantenating all files
    directory
        directory with the files. All files should
        be of the same structure and there should
        be no sub-directory inside it
    function
        function to be run on each file
        By default, pandas read_csv function is
        run on each file. If you specify your own
        function, it should have only filename
        as its argument and must return a dataframe
    kwargs
        kwargs for the pandas read_csv function
    """
    collect = []
    for root, directory, files in os.walk(directory):
        for file in files:
            filename = os.path.join(root, file)
            if function is None:
                temp = pd.read_csv(filename, **kwargs)
            else:
                temp = function(filename)
            collect.append(temp)
    result = pd.concat(collect).reset_index(drop=True)
    return result


import unittest
import pandas as pd
import context

from fastbt.datasource import DataSource

class TestDataSource(unittest.TestCase):

    def setUp(self):        
        df = pd.read_csv('tests/data/sample.csv', parse_dates=['timestamp'])
        self.ds = DataSource(data=df)

    def test_data(self):
        self.assertEqual(self.ds.data.iloc[20,1], 'five')
        self.assertEqual(self.ds.data.iloc[14,3], 112)
        self.assertEqual(self.ds.data.iloc[24,7], 10.54)

    def test_data_without_sort(self):
        df = pd.read_csv('tests/data/sample.csv', parse_dates=['timestamp'])
        self.ds = DataSource(data=df, sort=False)
        self.assertEqual(self.ds.data.iloc[9,4], 999)
        self.assertEqual(self.ds.data.iloc[24,6], 41688)
        self.assertEqual(self.ds.data.at[4, 'close'], 10.6)

    def test_initialize_case(self):
        df = pd.read_csv('tests/data/sample.csv', parse_dates=['timestamp'])
        df.columns = [x.upper() for x in df.columns]
        self.assertEqual(df.columns[0], 'TIMESTAMP')
        self.ds = DataSource(data=df)
        self.assertEqual(self.ds.data.columns[0], 'timestamp')


    def test_initialize_column_rename(self):
        df = pd.read_csv('tests/data/sample.csv', parse_dates=['timestamp'])
        df.columns = ['TS', 'TRADINGSYMBOL', 'OPEN', 'HIGH', 'LOW',
                    'CLOSE', 'VOLUME', 'PREVCLOSE']
        self.ds = DataSource(data=df, timestamp='TS', symbol='TRADINGSYMBOL')
        self.assertEqual(self.ds.data.columns[0], 'timestamp')
        self.assertEqual(self.ds.data.columns[1], 'symbol')

    def test_add_lag(self):
        length = len(self.ds.data)
        idx = pd.IndexSlice
        self.ds.add_lag(on='close')
        self.ds.add_lag(on='volume', period=2)
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        self.assertEqual(d.at[idx['2018-01-04', 'one'], 'lag_close_1'], 11)
        self.assertEqual(d.at[idx['2018-01-06', 'six'], 'lag_volume_2'], 86014)
        self.assertEqual(len(self.ds.data.columns), 10)
        self.assertEqual(len(self.ds.data), length)

    def test_add_lag_column_rename(self):
        idx = pd.IndexSlice
        self.ds.add_lag(on='close')
        self.ds.add_lag(on='close', col_name='some_col')
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        self.assertEqual(d.at[idx['2018-01-04', 'one'], 'lag_close_1'], 11)
        self.assertEqual(d.at[idx['2018-01-04', 'one'], 'some_col'], 11)
        self.assertEqual(d.at[idx['2018-01-05', 'three'], 'some_col'], 109)

    def test_add_pct_change(self):
        idx = pd.IndexSlice
        self.ds.add_pct_change(on='close')
        self.ds.add_pct_change(on='close', period=2)
        self.ds.add_pct_change(on='close', period=2, col_name='new_col')
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        R = lambda x: round(x,2)
        self.assertEqual(R(d.at[idx['2018-01-05', 'three'], 'chg_close_1']), -0.07)
        self.assertEqual(R(d.at[idx['2018-01-06', 'five'], 'chg_close_1']), 0.17)
        self.assertEqual(R(d.at[idx['2018-01-05', 'four'], 'chg_close_2']), 0.05)
        self.assertEqual(R(d.at[idx['2018-01-05', 'four'], 'new_col']), 0.05)
        self.assertEqual(R(d.at[idx['2018-01-03', 'six'], 'new_col']), -0.1)
        self.assertEqual(pd.isna(d.at[idx['2018-01-02', 'one'], 'new_col']), True)
        self.assertEqual(len(self.ds.data.columns), 11)

    def test_add_pct_change_lag(self):
        idx = pd.IndexSlice
        self.ds.add_pct_change(on='close', period=2, lag=1)
        self.ds.add_pct_change(on='close', period=1, lag=2)
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        R = lambda x: round(x,2)
        self.assertEqual(R(d.at[idx['2018-01-04', 'four'], 'chg_close_2']), 0.09)
        self.assertEqual(R(d.at[idx['2018-01-04', 'four'], 'chg_close_1']), 0.01)
        self.assertEqual(R(d.at[idx['2018-01-06', 'three'], 'chg_close_1']), -0.01)


    def test_add_pct_change_lag_col_name(self):
        idx = pd.IndexSlice
        self.ds.add_pct_change(on='high', period=2, lag=1)
        self.ds.add_pct_change(on='close', period=1, lag=2, col_name='lagged_2')
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        R = lambda x: round(x,2)
        self.assertEqual(R(d.at[idx['2018-01-05', 'six'], 'chg_high_2']), -0.04)
        self.assertEqual(R(d.at[idx['2018-01-04', 'four'], 'lagged_2']), 0.01)

    def test_formula_add_col_name(self):
        idx = pd.IndexSlice
        self.ds.add_formula('open+close', 'new_col')
        self.ds.add_formula('volume/close', 'new_col_2')
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        R = lambda x: round(x,2)
        self.assertEqual(R(d.at[idx['2018-01-04', 'four'], 'new_col']), 336)
        self.assertEqual(R(d.at[idx['2018-01-06', 'one'], 'new_col_2']), 77755.77)

    def test_formula_case_insensitive(self):
        idx = pd.IndexSlice
        self.ds.add_formula('OPEN+CLOSE', 'new_col')
        self.ds.add_formula('volume/close', 'NEW_COL_2')
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        R = lambda x: round(x,2)
        self.assertEqual(R(d.at[idx['2018-01-04', 'four'], 'new_col']), 336)
        self.assertEqual(R(d.at[idx['2018-01-06', 'one'], 'new_col_2']), 77755.77)


    def test_formula_calculated_column(self):
        idx = pd.IndexSlice
        self.ds.add_formula('(open+close)*100', 'new_col_1')
        self.ds.add_formula('volume/100', 'new_col_2')
        self.ds.add_formula('new_col_1+new_col_2', 'new_col_3')
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        R = lambda x: round(x,2)
        self.assertEqual(R(d.at[idx['2018-01-06', 'one'], 'new_col_3']), 10190.6)
        self.assertEqual(R(d.at[idx['2018-01-05', 'two'], 'new_col_3']), 200389.97)

    def test_rolling_simple(self):
        from pandas import isna
        q = 'symbol == "one"'
        df = pd.read_csv('tests/data/sample.csv', parse_dates=['timestamp']).query(q)
        df['r2'] = df['close'].rolling(2).mean()
        self.ds.add_rolling(2, col_name='r2')
        df2 = self.ds.data.query(q)
        print('RESULT' , df['r2'], df2['r2'])
        for a,b in zip(df['r2'], df2['r2']):
            if not(isna(a)):
                assert a==b     

    def test_rolling_values(self):
        idx = pd.IndexSlice
        self.ds.add_rolling(4, on='volume', function='max')
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        R = lambda x: round(x,2)
        self.assertEqual(d.at[idx['2018-01-05', 'five'], 'rol_max_volume_4'], 971704)
        self.assertEqual(d.at[idx['2018-01-05', 'six'], 'rol_max_volume_4'], 195539)
        self.assertEqual(d.at[idx['2018-01-04', 'three'], 'rol_max_volume_4'], 433733)
        # Adding lag and testing
        self.ds.add_rolling(4, on='volume', function='max', lag=1)
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        self.assertEqual(d.at[idx['2018-01-06', 'five'], 'rol_max_volume_4'], 971704)
        self.assertEqual(d.at[idx['2018-01-06', 'six'], 'rol_max_volume_4'], 195539)
        self.assertEqual(d.at[idx['2018-01-05', 'three'], 'rol_max_volume_4'], 433733)
        # Testing for 2 lags and column name
        self.ds.add_rolling(4, on='volume', function='max', lag=2, col_name='check')
        d = self.ds.data.set_index(['timestamp', 'symbol'])
        self.assertEqual(d.at[idx['2018-01-06', 'three'], 'check'], 433733) 

    def test_batch(self):
        length = len(self.ds.data)
        batch = [
            {'P': {'on': 'close', 'period': 1, 'lag': 1}},
            {'L': {'on': 'volume', 'period': 1}},
            {'F': {'formula': '(open+close)/2', 'col_name': 'AvgPrice'}},
            {'I': {'indicator': 'SMA', 'period': 3, 'lag': 1, 'col_name': 'SMA3'}},
            {'F': {'formula': 'avgprice + sma3', 'col_name': 'final'}},
            {'R': {'window': 3, 'function': 'mean'}}
        ]
        d = self.ds.batch_process(batch).set_index(['timestamp', 'symbol'])
        self.assertEqual(len(d.columns), 12)
        self.assertEqual(len(self.ds.data.columns), 14)
        self.assertEqual(len(self.ds.data), length)

    def test_raise_error_if_not_dataframe(self):
        pass


"""
Simple functional backtesting framework
based on end of day data

Assumptions
    1. Trades are done on an intraday basis
    2. Orders are placed at the beginning of the day
    3. Orders are closed when stop loss is hit or at the end of the day
    4. No target or pre-closing of orders
    5. All orders are placed/closed at the same time
    6. Calculations are based on simple returns
"""

import pandas as pd
from fastbt.datasource import DataSource

def tick(price, tick_size=0.05):
    """
    Rounds a given price to the requested tick
    """
    return round(price / tick_size)*tick_size

def isPrice(price, high, low):
    """
    Check whether the price is within the bound
    """
    return (price >= low) and (price <= high)

def fetch_data(universe='all', start=None, end=None, connection=None, tablename=None, where_clause = None):
    """
    Fetch data from SQL database
    where_clause
    additional where clauses as a list
    TO DO: Should adjust for date and time
    """
    from datetime import datetime, timedelta
    if end is None:
        end = datetime.today().strftime('%Y-%m-%d')
    if start is None:
        start = (datetime.today() - timedelta(days=30)).strftime('%Y-%m-%d')
    q = []
    select = "SELECT * from {tablename} where ".format(tablename=tablename)
    if universe != 'all':
        q.append("symbol in {universe}")
    q.append("timestamp >= '{start}'")
    q.append("timestamp <= '{end}'")
    if where_clause:
        [q.append(x)for x in where_clause]
    order_by = ' ORDER BY timestamp'
    query = ' AND '.join(q).format(universe=tuple(universe), 
        start=start, end=end)
    query = select + query + order_by
    # This should be any column
    data = pd.read_sql_query(query, connection, parse_dates=['timestamp'])
    # Delete index column if any
    if 'index' in data.columns:
        del data['index']
    return data

def prepare_data(data, columns=None):
    if columns:
        ds = DataSource(data) 
        return ds.batch_process(columns)
    else:
        return data

def apply_prices(data, conditions=None, price='open', stop_loss=0, order='B'):
    """
    Filter conditions and apply prices
    data
        datasource object
    conditions
        list of conditions as string
    price
        price at which order is to be booked
        as a formula string
    stop_loss
        stop loss as percentage from price
    order
        whether the order is Buy or Sell
        accepted values B or S
    """
    if order.upper() == 'B':
        multiplier = 1 - (stop_loss * 0.01)
    elif order.upper() == 'S':
        multiplier = 1 + (stop_loss * 0.01)
    else:
        raise ValueError('Order should be either B or S')

    # All conditions are forced to lower case AND ed
    if conditions:
        big_condition = '&'.join(['(' + c.lower() + ')' for c in conditions])
        # TO DO: Deal with Memory Error in case of lots of conditions
        # f shorthand for filtered
        f = data.query(big_condition).copy()
    else:
        f = data.copy()
    f['price'] = f.eval(price).apply(tick)
    f['stop_loss'] = (f['price'] * multiplier).apply(tick)

    # Price map to determine whether buy or sell is the entry
    p_map = {
        'B': ('buy', 'sell'),
        'S': ('sell', 'buy')
    }

    col_price, col_sl = p_map.get(order)

    f.loc[:, col_price] = [isPrice(p,h,l) for p,h,l in zip(
        f.price, f.high, f.low)] * f['price']
    f.loc[:, col_sl] = [isPrice(p,h,l) for p,h,l in zip(
        f.stop_loss, f.high, f.low)] * f['stop_loss']
    f.loc[f.buy == 0, 'buy'] = f.loc[f.buy == 0, 'close']
    f.loc[f.sell == 0, 'sell'] = f.loc[f.sell == 0, 'close']
    return f

def run_strategy(data, sort_by='price', sort_mode=True, limit=5, strategy=None):
    """
    Strategy to apply for each time bar    
    By default, NA's are dropped
    """
    grouped = data.dropna().groupby('timestamp')
    collect = []
    for name, group in grouped:
        if strategy:
            temp = group.apply(strategy)
        else:
            temp = group.sort_values(by=sort_by, ascending=sort_mode).iloc[:limit]
        collect.append(temp)
    return pd.concat(collect)

def get_output(data, capital=100000, leverage=1, commission=0, slippage=0):
    """
    By default, NA's are dropped
    """
    total_capital = capital * leverage
    df = data
    df['cnt'] = df.groupby('timestamp')['symbol'].transform(
        lambda x: len(x))
    df['qty'] = (total_capital/df['cnt']/df['price']).round()
    df['profit'] = df.eval('(sell-buy)*qty')
    df['commission'] = df.eval('(sell+buy)*qty') * commission * 0.01
    df['slippage'] = df.eval('(sell+buy)*qty') *  slippage * 0.01
    df['net_profit'] = df.eval('profit - commission - slippage')
    return df

def metrics(data, capital, benchmark=None):
    """
    Don't use this
    This is just to check results
    """
    if benchmark is None:
        benchmark = 0.08
    grouped = data.groupby('timestamp')
    cols = ['profit', 'commission', 'slippage', 'net_profit']
    profit = data.profit.sum()
    commission = data.commission.sum()
    slippage = data.slippage.sum()
    net_profit = data.net_profit.sum()
    high = grouped.net_profit.sum().cumsum().max()
    low = grouped.net_profit.sum().cumsum().min()
    drawdown = low/capital
    returns = net_profit/capital
    ret = grouped.agg({'profit': sum})
    sharpe = (ret.mean()/ret.std())[0]
    return {
        'profit': profit,
        'commission': commission,
        'slippage': slippage,
        'net_profit': net_profit,
        'high': high,
        'low': low,
        'drawdown': drawdown,
        'returns': returns,
        'sharpe': sharpe
    }       

def backtest(start=None, end=None,
            capital=100000, leverage=1, commission=0,
            slippage=0, price='open', stop_loss=0, order='B',
            universe='all', limit=5, columns=None, conditions=None,
            sort_by='price', sort_mode=True,
            connection=None, tablename=None,
            where_clause=None, data=None,
            strategy=None, output=None):
    """
    run the backtest
    start
        start date for the backtest
    end
        end date for the backtest
    capital
        capital to be invested
    strategy
        strategy should return a dataframe for each group
    """    
    if (data is None) and ((tablename is None) or (connection is None)):
        raise ValueError('No proper data source.\nEither specify a database connection and tablename or provide a dataframe')

    if data is None:
        data = fetch_data(universe=universe, start=start, end=end,
                     connection=connection, tablename=tablename,
                     where_clause=where_clause)

    # Check whether any data is available
    isNotEmpty = lambda x: True if len(data) > 0 else False

    if isNotEmpty(data):
        data = prepare_data(data, columns) 
        final = apply_prices(data, conditions, price, stop_loss, order) 
    else:
        raise ValueError('No data fetched from database')

    if isNotEmpty(final):   
        result = run_strategy(final, sort_by, sort_mode, limit, strategy)
    else:
        raise ValueError('No data after filtering all conditions')

    if output:
        return output(result)
    else:
        return get_output(result, capital, leverage, commission, slippage)

def backtest_from_excel(filename, data=None, connection=None, tablename=None):
    """
    Run a backtest from an excel file.
    The excel file should be prepared in the given template.
    filename
        excel filename along with full path
    data
        dataframe for backtest
    connection
        a SQL Alchemy connection string
    tablename
        SQL tablename
    """

    xls = pd.ExcelFile(filename)
    def parse_input_columns():
        # Parse and convert excel input for columns
        f_map = {
            'lag': 'L',
            'percent_change': 'P',
            'rolling': 'R',
            'formula': 'F',
            'indicator': 'I'
        }

        p_map = {
            'formula': 'formula',
            'indicator': 'indicator',
            'rolling': 'window'
        }

        records = xls.parse('columns').to_dict(orient='records')
        # Remove NaN's
        R = [{k:v for k,v in item.items() if pd.notnull(v)} for item in records]
        new_list = []
        for r in R:
            dct = {}
            code = f_map[r['col_type']]
            if p_map.get(r['col_type']):
                r[p_map[r['col_type']]] = r['argument'] 
                del r['argument']
            del r['col_type']
            dct[code] = r
            new_list.append(dct)
        return new_list

    params = {}
    parameters = xls.parse('parameters', header=None).set_index(0).to_dict()[1]
    params.update(parameters)

    columns = parse_input_columns()
    params.update({'columns': columns})

    conditions = list(xls.parse('conditions', header=None)[0])
    params.update({'conditions': conditions})

    params.update({'data': data, 'connection': connection, 'tablename': tablename})
    return backtest(**params)

def backtest_from_json(filename, data=None, connection=None, tablename=None):
    """
    Run a backtest from a JSON file
    JSON file should be in the expected format
    filename
        JSON filename
    """
    import json
    with open(filename, 'r') as f:
        params = json.load(f)
    params.update({'data': data, 'connection': connection, 'tablename': tablename})
    return backtest(**params)

def backtest_from_yaml(filename, data=None, connection=None, tablename=None):
    """
    Run a backtest from a YAML file
    YAML file should be in the expected format
    filename
        YAML filename
    """
    import yaml
    with open(filename, 'r') as f:
        params = yaml.load(f)
    params.update({'data': data, 'connection': connection, 'tablename': tablename})
    return backtest(**params)

def main():
    print('Hi')

if __name__ == "__main__":
    main()


"""
DataStore object with common functions
This is intended as a shortcut for otherwise
more verbose functions

Requirements
1) Input type should be a pandas dataframe
2) Input should have both symbol and timestamp columns
3) Indexes are reset and existing index dropped. So, if symbol or
   timestamp is in the index, convert them into columns  
4) All columns are renamed to lower case
"""

class DataSource(object):
    def __init__(self, data, symbol=None, timestamp=None, sort=True):
        """
        Initialize the dataframe
        By default, all columns would be converted into lower case, 
        values would be sorted by timestamp and indexes would be reset
        symbol
            symbol column
        timestamp
            timestamp column
        sort
            whether to sort data by timestamp
            if data is already sorted, pass False to save time
        """
        import hashlib
        self.hash = hashlib.sha1().hexdigest()
        self._isTALIB = True
        data = data.rename({symbol: 'symbol', timestamp: 'timestamp'},
            axis='columns')
        data = data.rename(str.lower, axis='columns')
        if sort:
            self._data = data.sort_values(by='timestamp').reset_index(drop=True)
        else:
            self._data = data.reset_index(drop=True)

        try:
            import talib
            self._func_map = {
                'BBANDS': (talib.BBANDS, ['close']),
                'DEMA': (talib.DEMA, ['close']),
                'EMA': (talib.EMA, ['close']),
                'KAMA': (talib.KAMA, ['close']),
                'MA': (talib.MA, ['close']),
                'MAMA': (talib.MAMA, ['close']),
                'MIDPOINT': (talib.MIDPOINT, ['close']),
                'MIDPRICE': (talib.MIDPRICE, ['high', 'low']),
                'SAR': (talib.SAR, ['high', 'low']),
                'SAREXT': (talib.SAREXT, ['high', 'low']),
                'SMA': (talib.SMA, ['close']),
                'STDDEV': (talib.STDDEV, ['close']),
                'TEMA': (talib.TEMA, ['close']),
                'TRIMA': (talib.TRIMA, ['close']),
                'WMA': (talib.WMA, ['close']),
                'AD': (talib.AD, ['high', 'low', 'close', 'volume']),
                'OBV': (talib.OBV, ['close', 'volume']),
                'ATR': (talib.ATR, ['high', 'low', 'close']),
                'NATR': (talib.NATR, ['high', 'low', 'close']),
                'TRANGE': (talib.TRANGE, ['high', 'low', 'close']),
                'ADX': (talib.ADX, ['high', 'low', 'close']),
                'AROON': (talib.AROON, ['high', 'low']),
                'BOP': (talib.BOP, ['open', 'high', 'low', 'close']),
                'CCI': (talib.CCI, ['high', 'low', 'close']),
                'DX': (talib.DX, ['high', 'low', 'close']),
                'MACD': (talib.MACD, ['close']),
                'MOM': (talib.MOM, ['close']),
                'ROC': (talib.ROC, ['close']),
                'RSI': (talib.RSI, ['close']),
                'STOCH': (talib.STOCH, ['high', 'low', 'close']),
                'STOCHF': (talib.STOCHF, ['high', 'low', 'close']),
                'STOCHRSI': (talib.STOCHRSI, ['close']),
                'ULTOSC': (talib.ULTOSC, ['high', 'low', 'close']),
                'WILLR': (talib.WILLR, ['high', 'low', 'close'])
            }
        except ModuleNotFoundError:
            self._isTALIB = False
            print('TALIB not installed')

    @property
    def data(self):
        return self._data

    def add_lag(self, on='close', period=1, col_name='auto'):
        """
        add lagged data based on symbol
        on
            column on which lag is to be added
        period
            period of lagging
        col_name
            column name
            By default, column name is created as lag_{{column}}_{{period}}
        """
        grouped = self.data.groupby('symbol')
        if col_name == 'auto':
            col_name = 'lag_' + on + '_' + str(period)
        col = grouped[on].transform(lambda x: x.shift(period))
        self._data[col_name.lower()] = col
        return self.data

    def add_pct_change(self, on='close', period=1, col_name='auto',
                        lag=None, **kwargs):
        """
        Add percentage change based on symbol
        lag
            also add lag by one period
        """
        grouped = self.data.groupby('symbol')
        if col_name == 'auto':
            col_name = 'chg_' + on + '_'+ str(period)
        col = grouped[on].transform(lambda x: x.pct_change(period, **kwargs))
        if lag:
            self._data[col_name + self.hash] = col
            self.add_lag(on=col_name+self.hash, period=lag, col_name=col_name)
            del self._data[col_name + self.hash]
        else:
            self._data[col_name.lower()] = col
        return self.data

    def add_rolling(self, window, groupby='symbol', on='close', 
                    function='mean', col_name='auto',lag=None, **kwargs):
        """
        Add rolling window statistics
        This is just a wrapper for the pandas rolling functions
        window
            rolling period window
        groupby
            column to group data by
        on
            column on which the rolling window is to be applied
        function
            function to be applied for the rolling window as
            a string. All pandas rolling window functions are
            accepted
        col_name
            column_name
        lag
            shift the data
        kwargs
            kwargs for the pandas window function
        This is just a wrapper for
        >>> df.groupby(groupby)[on].transform(lambda x: x.rolling(**kwargs).agg(function))
        """
        grouped = self.data.groupby(groupby)
        if col_name == 'auto':
            col_name = 'rol_{f}_{on}_{w}'.format(f=function, on=on, w=window)
        col = grouped[on].transform(lambda x: x.rolling(window, **kwargs).agg(function))
        if lag:
            self._data[col_name + self.hash] = col
            self.add_lag(on=col_name+self.hash, period=lag, col_name=col_name)
            del self._data[col_name + self.hash]
        else:
            self._data[col_name.lower()] = col
        return self.data
        
    def add_formula(self, formula, col_name):
        """
        Add a formula to the dataframe
        Formulas are simple one-line expressions
            see numexpr expressions for more details
        formula
            formula as a string
        col_name
            column name to use
        Note
        -----
        Formula string is renamed to lower case.
        So, formula, FORMULA, Formula are all equivalent
        """
        self._data[col_name.lower()] = self._data.eval(formula.lower())
        return self.data

    def add_indicator(self, indicator, period=None, col_name='auto', 
                    lag=None, **kwargs):
        """
        Add an indicator
        indicator
            indicator name as a string - case insensitive
        period
            the timeperiod argument for talib library
        col_name
            column name for this indicator
        kwargs for the indicator
        """
        if not(self._isTALIB):
            return 'TALIB not installed'
        indicator = indicator.upper()
        if period is not None:
            kwargs.update({'timeperiod': period})
        func, args = self._func_map.get(indicator, (None, None))

        if func is None:            
            raise NameError('Function not available\n' + 
             'Only the following functions are available\n{}'.format(self._func_map.keys()))

        def apply_func(group):
            cols = [group[arg] for arg in args]
            return func(*cols, **kwargs)

        data = self.data.reset_index(drop=True)
        grouped = data.groupby('symbol')
        result = grouped.apply(apply_func).reset_index(level='symbol')
        if col_name == 'auto':
            col_name = indicator + '_' + str(period if period is not None else 0)
        del result['symbol'] # Removed for easy join
        if lag:
            result.columns = [col_name + self.hash]
            self._data = data.join(result)
            self.add_lag(on=col_name+self.hash, period=lag, col_name=col_name)
            del self._data[col_name + self.hash]
        else:
            result.columns = [col_name.lower()]
            self._data = data.join(result)
        return self.data

    def batch_process(self, batch):
        """
        Process data in batch
        batch
            list of key,value pairs as arguments
            with key name being the code and
            value being a dictionary of arguments
            to the function

        code
        -----
        L - lag
        P - percent change
        F - formula
        I - indicator 
        """
        f_map = {
        'L': self.add_lag,
        'P': self.add_pct_change,
        'F': self.add_formula,
        'I': self.add_indicator,
        'R': self.add_rolling
        }

        for item in batch:
            # Parse data arguments and return final data
            k,v = tuple(item.items())[0]
            func = f_map[k]
            self._data = func(**v)        
        return self.data


import pandas as pd
from sqlalchemy import create_engine
import unittest
import os
import shutil
import tempfile
from numpy import dtype
import pytest
from math import isclose
from random import randint
import context

from fastbt.loaders import DataLoader, apply_adjustment, collate_data

def compare(frame1, frame2):
    """
    Compare a random value from 2 dataframes
    return
        True if values are equal else False
    """
    r1 = randint(0, len(frame1)-1)
    r2 = randint(0, len(frame1.columns) - 1)
    return frame1.iloc[r1, r2] == frame2.iloc[r1, r2]


class TestLoader(unittest.TestCase):

	def test_create_hdf_file(self):
		with tempfile.NamedTemporaryFile() as fp:		
			dl = DataLoader('tests/data/eoddata', engine=fp.name,
				mode='HDF', tablename='eod')
			dl.load_data()
			self.assertEqual(len(pd.read_hdf(fp.name, 'data/eod')), 10030)
			self.assertEqual(len(pd.read_hdf(fp.name, 'updated/eod')), 5)				

	def test_create_database(self):
		engine = create_engine('sqlite://')
		dl = DataLoader('tests/data/eoddata', engine=engine, 
			mode='SQL', tablename='eod')
		dl.load_data()
		self.assertEqual(len(pd.read_sql_table('eod', engine)), 10030)
		self.assertEqual(len(pd.read_sql_table('updated_eod', engine)), 5)

	def test_run_loader_multiple_times(self):
		with tempfile.NamedTemporaryFile() as fp:
			dl = DataLoader('tests/data/eoddata', engine=fp.name,
			mode='HDF', tablename='eod')
			for i in range(5):
				dl.load_data()
			engine = create_engine('sqlite://')
			dl = DataLoader('tests/data/eoddata', engine=engine, 
				mode='SQL', tablename='eod')
			for i in range(5):
				dl.load_data()
			shape_hdf = len(pd.read_hdf(fp.name, 'data/eod'))
			shape_sql = len(pd.read_sql_table('eod', engine))
			self.assertEqual(shape_hdf, shape_sql)
			self.assertEqual(shape_hdf, 12053)

	def test_existing_hdf_file(self):
		with tempfile.NamedTemporaryFile() as fp:
			dl = DataLoader('tests/data/eoddata', engine=fp.name,
				mode='HDF', tablename='eod')
			shutil.copy2('tests/data/eoddata/INDEX_20180731.txt', 
						'tests/data/eoddata/INDEX_20000000.txt')
			dl.load_data()
			self.assertEqual(len(pd.read_hdf(fp.name, 'data/eod')), 12053)
			self.assertEqual(len(pd.read_hdf(fp.name, 'updated/eod')), 6)		


	def test_existing_database(self):
		engine = create_engine('sqlite://')
		dl = DataLoader('tests/data/eoddata', engine=engine, 
			mode='SQL', tablename='eod')
		shutil.copy2('tests/data/eoddata/INDEX_20180731.txt', 
					 'tests/data/eoddata/INDEX_20000000.txt')
		dl.load_data()
		self.assertEqual(len(pd.read_sql_table('eod', engine)), 12053)
		self.assertEqual(len(pd.read_sql_table('updated_eod', engine)), 6)


	def test_wrong_mode(self):
		with tempfile.NamedTemporaryFile() as fp:
			dl = DataLoader('tests/data/eoddata', engine=fp.name,
				mode='SQL', tablename='eod')
			with self.assertRaises(Exception):
				dl.load_data()

		with self.assertRaises(TypeError):
			DataLoader('tests/data/eoddata', engine='some_random_mode',
				mode='CSV', tablename='eod')


	@classmethod
	def tearDownClass(self):
		os.remove('tests/data/eoddata/INDEX_20000000.txt')

# rename columns
rename =  {
	'<ticker>': 'symbol',
	'<date>': 'date',
	'<open>': 'open',
	'<high>': 'high',
	'<low>': 'low',
	'<close>': 'close',
	'<vol>': 'vol'
	}

def test_HDF_rename_columns():
	with tempfile.NamedTemporaryFile() as fp:
		dl = DataLoader('tests/data/eoddata', engine=fp.name,
			mode='HDF', tablename='eod')
		dl.load_data(columns=rename)
		df = pd.read_hdf(fp.name, 'data/eod')
		assert len(df) == 10030
		assert len(pd.read_hdf(fp.name, 'updated/eod')) == 5
		cols = ['symbol', 'date', 'open', 'high', 'low', 'close', 'vol']
		for x,y in zip(df.columns, cols):
			assert x == y

def test_SQL_rename_columns():
	engine = create_engine('sqlite://')
	dl = DataLoader('tests/data/eoddata', engine=engine, 
		mode='SQL', tablename='eod')
	dl.load_data(columns=rename)
	df = pd.read_sql_table('eod', engine)
	assert len(df) == 10030
	cols = ['symbol', 'date', 'open', 'high', 'low', 'close', 'vol']
	for x,y in zip(df.columns, cols):
		assert x == y

def test_HDF_parse_dates():
	with tempfile.NamedTemporaryFile() as fp: 
		dl = DataLoader('tests/data/eoddata', engine=fp.name,
			mode='HDF', tablename='eod')
		dl.load_data(columns=rename, parse_dates=['<date>'])
		df = pd.read_hdf(fp.name, 'data/eod')
		assert df.dtypes['date'] == dtype('<M8[ns]')

def test_SQL_parse_dates():
	engine = create_engine('sqlite://')
	dl = DataLoader('tests/data/eoddata', engine=engine, 
		mode='SQL', tablename='eod')
	dl.load_data(columns=rename, parse_dates=['<date>'])
	df = pd.read_sql_table('eod', engine)
	assert df.dtypes['date'] == dtype('<M8[ns]')

def test_HDF_parse_dates_auto():
	with tempfile.NamedTemporaryFile() as fp:
		dl = DataLoader('tests/data/eoddata', engine=fp.name,
			mode='HDF', tablename='eod')
		dl.load_data(columns=rename)
		df = pd.read_hdf(fp.name, 'data/eod')
		assert df.dtypes['date'] == dtype('<M8[ns]')

def test_SQL_parse_dates_auto():
	engine = create_engine('sqlite://')
	dl = DataLoader('tests/data/eoddata', engine=engine, 
		mode='SQL', tablename='eod')
	dl.load_data(columns=rename)
	df = pd.read_sql_table('eod', engine)
	assert df.dtypes['date'] == dtype('<M8[ns]')

def test_HDF_post_func():
	with tempfile.NamedTemporaryFile() as fp:
		dl = DataLoader('tests/data/eoddata', engine=fp.name,
			mode='HDF', tablename='eod')
		def add_filename(x,y,z):
			x['filename'] = y
			x['avgprice'] = (x['open'] + x['close'])/2
			return x
		dl.load_data(columns=rename, postfunc=add_filename)
		df = pd.read_hdf(fp.name, 'data/eod')
		assert df.dtypes['date'] == dtype('<M8[ns]')
		assert df.shape[1] == 9
		assert 'filename' in df.columns
		assert 'avgprice' in df.columns

def test_SQL_post_func():
	engine = create_engine('sqlite://')
	dl = DataLoader('tests/data/eoddata', engine=engine, 
		mode='SQL', tablename='eod')
	def add_filename(x,y,z):
		x['filename'] = y
		x['avgprice'] = (x['open'] + x['close'])/2
		return x
	dl.load_data(columns=rename, postfunc=add_filename)
	df = pd.read_sql_table('eod', engine)
	assert df.dtypes['date'] == dtype('<M8[ns]')
	assert df.shape[1] == 9
	assert 'filename' in df.columns
	assert 'avgprice' in df.columns

def test_apply_adj_mul():
	df = pd.read_csv('tests/data/BTC.csv', parse_dates=['date'])
	adj_df = apply_adjustment(df, adj_date='2018-07-21', 
			adj_value=1/2)
	adj_df = adj_df.set_index('date').sort_index()
	cols = ['open', 'high', 'low', 'close']
	assert adj_df.loc['2018-07-11', 'open'] == 3151.24
	assert adj_df.loc['2018-07-21', 'close'] == 3702.15
	for a,b in zip(adj_df.loc['2018-07-21', cols] ,
		(3665.27, 3721.2, 3608.47, 3702.15)):
		assert a == b
	assert adj_df.loc['2018-08-01', 'high'] == 15500.16
	assert adj_df.loc['2018-07-11', 'volume'] == 2481016

def test_apply_adj_sub():
	df = pd.read_csv('tests/data/BTC.csv', parse_dates=['date'])
	adj_df = apply_adjustment(df, adj_date='2018-08-01', 
			adj_value=100, adj_type='sub')
	adj_df = adj_df.set_index('date').sort_index()
	assert adj_df.loc['2018-07-16', 'close'] == 6626.4
	assert adj_df.loc['2018-08-01', 'high'] == 15500.16
	assert adj_df.loc['2018-08-10', 'close'] == 12286.6

def test_apply_adj_sub_negative():
	df = pd.read_csv('tests/data/BTC.csv', parse_dates=['date'])
	adj_df = apply_adjustment(df, adj_date='2018-08-01', 
			adj_value=-100, adj_type='sub')
	adj_df = adj_df.set_index('date').sort_index()
	assert adj_df.loc['2018-07-16', 'close'] == 6826.4
	assert adj_df.loc['2018-08-01', 'high'] == 15500.16
	assert adj_df.loc['2018-08-10', 'close'] == 12286.6

def test_apply_adj_raise_error():
	with pytest.raises(ValueError):
		df = pd.read_csv('tests/data/BTC.csv', parse_dates=['date'])
		adj_df = apply_adjustment(df, adj_date='2018-08-01', 
			adj_value=-100, adj_type='div')

def test_apply_adj_date_col():
	df = pd.read_csv('tests/data/BTC.csv', parse_dates=['date'])
	df['timestamp'] = df['date']
	del df['date']
	adj_df = apply_adjustment(df, adj_date='2018-07-21', 
			adj_value=1/2, date_col='timestamp')
	adj_df = adj_df.set_index('timestamp').sort_index()
	assert adj_df.loc['2018-07-11', 'open'] == 3151.24
	assert adj_df.loc['2018-07-21', 'close'] == 3702.15

def test_apply_adj_cols():
	df = pd.read_csv('tests/data/BTC.csv', parse_dates=['date'])
	adj_df = apply_adjustment(df, adj_date='2018-07-21', 
			adj_value=1/2, cols=['open', 'high'])
	adj_df = adj_df.set_index('date').sort_index()
	cols = ['open', 'high', 'low', 'close']
	assert adj_df.loc['2018-07-11', 'open'] == 3151.24
	assert adj_df.loc['2018-07-11', 'close'] == 6381.87

def test_apply_split_SQL_dataloader():
	engine = create_engine('sqlite://')
	dl = DataLoader(directory='tests/data/NASDAQ/data', mode='SQL',
		engine=engine, tablename='eod')
	dl.load_data()
	dl.apply_splits(directory='tests/data/NASDAQ/adjustments/')
	df = pd.read_sql_table('eod', engine)
	result = pd.read_csv('NASDAQ/nasdaq_results.csv', parse_dates=['date'])
	splits = pd.read_csv('tests/data/NASDAQ/adjustments/splits.csv',
		parse_dates=['date'])
	for i, row in splits.iterrows():
		sym = row.at['symbol']
		cond = 'symbol == "{}"'.format(sym)
		frame1 = df.query(cond).sort_values(by='date').reset_index(drop=True)
		frame2 = result.query(cond).sort_values(by='date').reset_index(drop=True)
		L = len(frame1)
		cols = frame1.columns
		for i in range(L):
			for j in cols:
				if j in ['open', 'high', 'low', 'close', 'volume']:
					a = frame1.loc[i,j]
					b = frame2.loc[i,j]
					assert isclose(a,b,abs_tol=0.015)
				else:
					assert frame1.loc[i,j] == frame2.loc[i,j]

def test_apply_split_SQL_dataloader():
	engine = create_engine('sqlite://')
	dl = DataLoader(directory='tests/data/NASDAQ/data', mode='SQL',
		engine=engine, tablename='eod')
	dl.load_data()
	dl.apply_splits(directory='tests/data/NASDAQ/adjustments/')
	df = pd.read_sql_table('eod', engine)
	result = pd.read_csv('tests/data/NASDAQ/nasdaq_results.csv', parse_dates=['date'])
	splits = pd.read_csv('tests/data/NASDAQ/adjustments/splits.csv',
		parse_dates=['date'])
	for i, row in splits.iterrows():
		sym = row.at['symbol']
		cond = 'symbol == "{}"'.format(sym)
		frame1 = df.query(cond).sort_values(by='date').reset_index(drop=True)
		frame2 = result.query(cond).sort_values(by='date').reset_index(drop=True)
		L = len(frame1)
		cols = frame1.columns
		for i in range(L):
			for j in cols:
				if j in ['open', 'high', 'low', 'close', 'volume']:
					a = frame1.loc[i,j]
					b = frame2.loc[i,j]
					assert isclose(a,b,abs_tol=0.015)
				else:
					assert frame1.loc[i,j] == frame2.loc[i,j]


def test_apply_split_HDF_dataloader():
	with tempfile.NamedTemporaryFile() as fp:
		engine = fp.name
		dl = DataLoader(directory='tests/data/NASDAQ/data', mode='HDF',
			engine=engine, tablename='eod')
		dl.load_data()
		dl.apply_splits(directory='tests/data/NASDAQ/adjustments/')
		df = pd.read_hdf(engine, 'data/eod')
		result = pd.read_csv('tests/data/NASDAQ/nasdaq_results.csv', parse_dates=['date'])
		splits = pd.read_csv('tests/data/NASDAQ/adjustments/splits.csv',
			parse_dates=['date'])
		for i, row in splits.iterrows():
			sym = row.at['symbol']
			cond = 'symbol == "{}"'.format(sym)
			frame1 = df.query(cond).sort_values(by='date').reset_index(drop=True)
			frame2 = result.query(cond).sort_values(by='date').reset_index(drop=True)
			L = len(frame1)
			cols = frame1.columns
			for i in range(L):
				for j in cols:
					if j in ['open', 'high', 'low', 'close', 'volume']:
						a = frame1.loc[i,j]
						b = frame2.loc[i,j]
						print(a,b,sym)
						assert isclose(a,b,abs_tol=0.015)
					else:
						assert frame1.loc[i,j] == frame2.loc[i,j]

def test_collate_data():
	df = collate_data('tests/data/NASDAQ/data', parse_dates=['Date'])
	df = df.rename(lambda x: x.lower(), axis='columns')
	df = df.sort_values(by=['date', 'symbol'])
	engine = create_engine('sqlite://')
	dl = DataLoader(directory='tests/data/NASDAQ/data', mode='SQL',
		engine=engine, tablename='eod')
	dl.load_data()
	df2 = pd.read_sql_table('eod', engine).sort_values(by=['date', 'symbol'])
	assert len(df) == len(df2)
	for i in range(100):
		assert compare(df, df2)

def test_collate_data_function():
	def f(x):
		return pd.read_csv(x).iloc[:10, :3]
	df = collate_data('tests/data/NASDAQ/data', function=f)
	assert len(df) == 80
	assert df.shape == (80, 3)


import pandas as pd
import itertools as it

def multi_args(function, constants, variables, isProduct=False, maxLimit=None):
    """
    Run a function on different parameters and
    aggregate results
    function
        function to be parametrized
    constants
        arguments that would remain constant
        throughtout all the scenarios
        dictionary with key being argument name
        and value being the argument value
    variables
        arguments that need to be varied
        dictionary with key being argument name
        and value being list of argument values
        to substitute
    isProduct
        list of variables for which all combinations
        are to be tried out.
    maxLimit
        Maximum number of simulations to be run
        before terminating. Useful in case of long
        running simulations.
        default 1000

    By default, this function zips through each of the
    variables but if you need to have the Cartesian
    product, specify those variables in isProduct.

    returns a Series with different variables and
    the results
    """
    from functools import partial
    import concurrent.futures

    if maxLimit:
        MAX_LIMIT = maxLimit
    else:
        MAX_LIMIT = 1000
        
    func = partial(function, **constants)
    arg_list = []
    if isProduct:
        args = it.product(*variables.values())
    else:
        args = zip(*variables.values())
    keys = variables.keys()
    with concurrent.futures.ProcessPoolExecutor() as executor:
        tasks = []
        for i, arg in enumerate(args):
            kwds = {a:b for a,b in zip(keys, arg)}
            tasks.append(executor.submit(func, **kwds))
            arg_list.append(arg)
            i += 1
            if i >= MAX_LIMIT:
                print('MAX LIMIT reached', MAX_LIMIT)
                break
    result = [task.result() for task in tasks] 
    s = pd.Series(result)
    s.name = 'values'
    s.index = pd.MultiIndex.from_tuples(arg_list, names=keys)
    return s

def stop_loss(price, stop_loss, order='B', tick_size=0.05):
    """
    Return the stop loss for the order
    price
        price from which stop loss is to be calculated
    stop_loss
        stop loss percentage from price
    order
        the original order type - B for Buy and S for Sell
        If the original order is buy, then a sell stop
        loss is generated and vice-versa
    tick_size
        tick_size to be rounded off
    >>> stop_loss(100, 3)
    >>> 97

    Notes
    ------
    * passing a negative value may throw unexpected results
    * raises ValueError if order is other than B or S

    """
    if order == 'B':
        return tick(price * (1 - stop_loss * 0.01), tick_size)
    elif order == 'S':
        return tick(price * (1 + stop_loss * 0.01), tick_size)
    else:
        raise ValueError('order should be either B or S')


def tick(price, tick_size=0.05):
    """
    Rounds a given price to the requested tick
    """
    return round(price / tick_size)*tick_size

def create_orders(data, rename, **kwargs):
    """
    create an orders dataframe from an existing dataframe
    by renaming columns and providing additional columns
    data
        dataframe
    rename
        columns to be renamed as dictionary
    kwargs
        key value pairs with key being column names
        and values being dataframe values
    """
    data = data.rename(rename, axis='columns')
    for k,v in kwargs.items():
        data[k] = v
    return data

def recursive_merge(dfs, on=None, how='inner',columns= {}):
    """
    Recursively merge all dataframes in the given list
    
    Given a list of dataframes, merge them based on index or columns.
    By default, dataframes are merged on index. Specify the **on**
    argument to merge by columns. The "on" columns should be available
    in all the dataframes

    Parameters
    -----------
    dfs
        list of dataframes
    on
        columns on which the dataframes are to be merged.
        By default, merge is done on index
    how
        how to apply the merge
        {'left', 'right', 'outer', 'inner'}, default 'inner'.
        Same as pandas merge
    columns
        To return only specific columns from specific dataframes,
        pass them as a dictionary with key being the index of the 
        dataframe in the list and value being the list of columns
        to merge. **your keys should be string**
        See examples for more details
        >>> recursive_merge(dfs, columns{'1': ['one', 'two']})
        Fetch only the columns one and two from the second dataframe
    """
    data = dfs[0]
    for i, d in enumerate(dfs[1:], 1):
        if columns.get(str(i)):
            cols = list(columns.get(str(i)))         
            cols.extend(on)            
        else:
            cols = d.columns

        if on is None:
            data = data.merge(d[cols], how=how, left_index=True, right_index=True)
        else:
            data = data.merge(d[cols], how=how, on=on)
    return data   
    
if __name__ == "__main__":
    pass

